@@ -2,6 +2,7 @@
 
 from .arrays import ArrayType
 from .beads import (
+    NEUTRAL_BETA,
     MethylData,
     RawData,
     ReferenceMethylData,
@@ -9,15 +10,9 @@ from .beads import (
     idat_basepaths,
     is_valid_idat_basepath,
 )
-from .cache import get_id_tuple, input_args_id, memoize
+from .cache import clear_cache, get_id_tuple, input_args_id, memoize
 from .chromosome import Chromosome
 from .cnv import CNV, ZIP_ENDING, Annotation, _get_cgsegment
-from .genetic_data import (
-    CHROMOSOME_DATA,
-    GAPS,
-    GENES,
-    IMPORTANT_GENES,
-)
 from .idat import IdatParser
 from .manifests import Manifest
 from .plots import (
@@ -31,15 +26,13 @@ from .probes import Channel, InfiniumDesignType, ProbeType
 __all__ = [
     "Annotation",
     "ArrayType",
-    "CHROMOSOME_DATA",
     "CNV",
     "CNVPlot",
+    "NEUTRAL_BETA",
     "input_args_id",
+    "clear_cache",
     "Channel",
     "Chromosome",
-    "GAPS",
-    "GENES",
-    "IMPORTANT_GENES",
     "IdatParser",
     "InfiniumDesignType",
     "Manifest",
