@@ -5,6 +5,7 @@ preprocessing techniques, normalization, and data handling.
 """
 
 import collections
+import os
 import pickle
 from functools import reduce
 from pathlib import Path
@@ -18,34 +19,35 @@ from mepylome.dtypes.cache import cache_key, input_args_id, memoize
 from mepylome.dtypes.idat import IdatParser
 from mepylome.dtypes.manifests import Manifest
 from mepylome.dtypes.probes import Channel, ProbeType
-from mepylome.utils.files import MEPYLOME_TMP_DIR
-from mepylome.utils.varia import normexp_get_xs
+from mepylome.utils.varia import MEPYLOME_TMP_DIR, normexp_get_xs
 
 ENDING_GRN = "_Grn.idat"
 ENDING_RED = "_Red.idat"
 ENDING_GZ = ".gz"
+ENDING_SUFFIXES = ("_Grn.idat", "_Red.idat", "_Grn.idat.gz", "_Red.idat.gz")
 
-NEUTRAL_BETA = 0.49
+NEUTRAL_BETA = 0.5
 
 
 def is_valid_idat_basepath(basepath):
     """Checks if the given basepath(s) point to valid IDAT files."""
     if not isinstance(basepath, list):
         basepath = [basepath]
+    basepath = [str(x) for x in basepath]
     return all(
         (
-            Path(str(x) + ENDING_GRN).exists()
-            or Path(str(x) + ENDING_GRN + ENDING_GZ).exists()
+            os.path.exists(x + ENDING_GRN)
+            or os.path.exists(x + ENDING_GRN + ENDING_GZ)
         )
         and (
-            Path(str(x) + ENDING_RED).exists()
-            or Path(str(x) + ENDING_RED + ENDING_GZ).exists()
+            os.path.exists(x + ENDING_RED)
+            or os.path.exists(x + ENDING_RED + ENDING_GZ)
         )
         for x in basepath
     )
 
 
-def idat_basepaths(files):
+def idat_basepaths(files, only_valid=False):
     """Returns unique basepaths from IDAT files or directory.
 
     This function processes a list of IDAT files or a directory containing IDAT
@@ -55,6 +57,8 @@ def idat_basepaths(files):
 
     Args:
         files (path or list): A file or directory path or a list of file paths.
+        only_valid (bool): If True, only returns basepaths that point to valid
+            IDAT file pairs. Defaults is 'False'.
 
     Returns:
         list: A list of unique basepaths corresponding to the provided IDAT
@@ -62,8 +66,8 @@ def idat_basepaths(files):
             considered.
 
     Example:
-        >>> idat_basepaths("/path/to/idat/dir")
-        [PosixPath('/path/to/idat/file1'), PosixPath('/path/to/idat/file2')]
+        >>> idat_basepaths("/path/to/dir")
+        [PosixPath('/path/to/dir/file1'), PosixPath('/path/to/dir/file2')]
 
         >>> idat_basepaths(["/path1/file1_Grn.idat", "/path2/file2_Red.idat"])
         [PosixPath('/path1/file1'), PosixPath('/path2/file2')]
@@ -71,25 +75,40 @@ def idat_basepaths(files):
         >>> idat_basepaths("/path/to/idat/file_Grn.idat.gz")
         [PosixPath('/path/to/idat/file')]
     """
-    if isinstance(files, list):
-        _files = files
-    elif Path(files).expanduser().is_dir():
-        # If basenames is dir take all files in it
-        _files = Path(files).expanduser().rglob("*_[GR][re][nd].idat*")
-    else:
-        _files = [files]
-    # Remove file endings
+
+    def get_idat_files(file_or_dir):
+        path = os.path.expanduser(file_or_dir)
+        # If path is dir take all files in it
+        if os.path.isdir(path):
+            for dirpath, _, filenames in os.walk(path, followlinks=True):
+                for filename in filenames:
+                    if filename.endswith(ENDING_SUFFIXES):
+                        yield os.path.join(dirpath, filename)
+        else:
+            yield path
+
+    def strip_suffix(file_path):
+        for suffix in ENDING_SUFFIXES:
+            if file_path.endswith(suffix):
+                return file_path[: -len(suffix)]
+        return file_path
+
+    if not isinstance(files, list):
+        files = [files]
     _files = [
-        Path(
-            str(name)
-            .replace(ENDING_GZ, "")
-            .replace(ENDING_GRN, "")
-            .replace(ENDING_RED, "")
-        ).expanduser()
-        for name in _files
+        strip_suffix(idat_file)
+        for file_or_dir in files
+        for idat_file in get_idat_files(file_or_dir)
     ]
     # Remove duplicates, keep ordering
-    return list(dict.fromkeys(_files))
+    unique_basepaths_dict = dict.fromkeys(_files)
+    if only_valid:
+        return [
+            Path(base)
+            for base in unique_basepaths_dict
+            if is_valid_idat_basepath(base)
+        ]
+    return [Path(base) for base in unique_basepaths_dict]
 
 
 def idat_paths_from_basenames(basenames):
@@ -915,14 +934,18 @@ class MethylData:
     def _get_beta(
         methylated, unmethylated, offset=0, beta_threshold=0, *, min_zero=True
     ):
-        assert offset >= 0, "offset must be non-negative"
-        assert (
-            0 <= beta_threshold <= 0.5
-        ), "beta_threshold must be between 0 and 0.5"
+        if offset < 0:
+            msg = "'offset' must be non-negative"
+            raise ValueError(msg)
+
+        if not (0 <= beta_threshold <= 0.5):
+            msg = "'beta_threshold' must be between 0 and 0.5"
+            raise ValueError(msg)
 
         if min_zero:
             methylated = np.maximum(methylated, 0)
             unmethylated = np.maximum(unmethylated, 0)
+
         # Ignore division by zero
         with np.errstate(divide="ignore", invalid="ignore"):
             betas = methylated / (methylated + unmethylated + offset)
@@ -1026,8 +1049,7 @@ class ReferenceMethylData:
         for idat_file in tqdm(
             idat_files, desc="Categorizing reference IDAT files"
         ):
-            raw_data = RawData(idat_file)
-            array_type = raw_data.array_type
+            array_type = ArrayType.from_idat(idat_file)
             reference_files[array_type].append(idat_file)
         for array_type, file_list in tqdm(
             reference_files.items(), desc="Processing reference IDAT files"
@@ -1045,14 +1067,14 @@ class ReferenceMethylData:
     @staticmethod
     def pickle_filename(prep, idat_files):
         return MEPYLOME_TMP_DIR / input_args_id(
-            "ReferenceMethylData", prep, sorted(str(x) for x in idat_files)
+            "Ref", prep, sorted(str(x) for x in idat_files)
         )
 
     def __getitem__(self, array_type):
         if array_type not in self._methyl_data:
             msg = (
                 f"No copy number neutral reference files found for "
-                f"array type {array_type.value}."
+                f"array type '{array_type.value}'."
             )
             raise ValueError(msg)
         return self._methyl_data[array_type]
