@@ -6,6 +6,8 @@ analysis.
 
 import heapq
 import io
+import logging
+import pickle
 import zipfile
 from functools import lru_cache
 from pathlib import Path
@@ -16,33 +18,48 @@ import pyranges as pr
 
 from mepylome.dtypes.arrays import ArrayType
 from mepylome.dtypes.beads import MethylData, ReferenceMethylData
-from mepylome.dtypes.cache import cache_key, memoize
+from mepylome.dtypes.cache import cache_key, input_args_id, memoize
 from mepylome.dtypes.chromosome import Chromosome
-from mepylome.dtypes.genetic_data import GAPS, GENES
 from mepylome.dtypes.manifests import Manifest
 from mepylome.dtypes.plots import CNVPlot
-from mepylome.utils.files import MEPYLOME_TMP_DIR, ensure_directory_exists
-from mepylome.utils.varia import log
+from mepylome.utils.files import ensure_directory_exists, get_resource_path
+from mepylome.utils.varia import CONFIG, MEPYLOME_TMP_DIR
 
-UNSET = object()
-ZIP_ENDING = "_cnv.zip"
+logger = logging.getLogger(__name__)
+
+ZIP_ENDING = CONFIG["suffixes"]["cnv_zip"]
+PACKAGE_DIR = get_resource_path("mepylome")
+
+# Data copied from conumee
+GAPS = PACKAGE_DIR / CONFIG["paths"]["gaps"]
+
+# HG19 Gene data downloaded from:
+# https://grch37.ensembl.org/biomart/martview
+GENES = PACKAGE_DIR / CONFIG["paths"]["genes"]
 
 
-def _missing_cbseg_msg():
-    log(
-        "[CNV] **Warning**: Segmentation won't be calculated due to missing "
-        "'linear_segment' resp. 'cbseg' package. See documentation."
-    )
+def _get_cgsegment():
+    try:
+        import ruptures
 
+        def function(bin_values):
+            algo = ruptures.KernelCPD("linear").fit(bin_values)
+            segments = algo.predict(pen=1)
+            return [
+                [start, end] for start, end in zip([0] + segments, segments)
+            ]
 
-def _get_cgsegment(verbose=False):
+        return function
+    except Exception:
+        pass
     try:
         import linear_segment
 
         def function(bin_values):
-            return linear_segment.segment(
+            segments = linear_segment.segment(
                 bin_values, shuffles=1000, cutoff=0.3
             )
+            return [[s.start, s.end] for s in segments]
 
         return function
     except Exception:
@@ -51,12 +68,16 @@ def _get_cgsegment(verbose=False):
         import cbseg
 
         def function(bin_values):
-            return cbseg.segment(bin_values, shuffles=1000, p=0.0001)
+            segments = cbseg.segment(bin_values, shuffles=1000, p=0.0001)
+            return [[s.start, s.end] for s in segments]
 
         return function
     except Exception:
-        if verbose:
-            _missing_cbseg_msg()
+        logger.warning(
+            "**Warning**: Segmentation won't be calculated due to missing "
+            "'linear_segment', 'cbseg' or 'ruptures' package. See "
+            "documentation"
+        )
         return None
 
 
@@ -76,8 +97,6 @@ class Annotation:
             Defaults to 50000.
         min_probes_per_bin (int, optional): The minimum number of probes
             per bin. Defaults to 15.
-        verbose (bool, optional): Whether to display verbose output during
-            annotation processing. Defaults to False.
 
     Attributes:
         manifest (Manifest): The manifest to use.
@@ -89,8 +108,6 @@ class Annotation:
             genes).
         bin_size (int): The base-pair size of the bins.
         min_probes_per_bin (int): The minimum number of probes per bin.
-        verbose (bool): Whether to display verbose output during annotation
-            processing.
         chromsizes (dict): Dictionary containing chromosome sizes.
     """
 
@@ -100,11 +117,10 @@ class Annotation:
         cls,
         manifest=None,
         array_type=None,
-        gap=UNSET,
-        detail=UNSET,
+        gap=None,
+        detail=None,
         bin_size=50000,
         min_probes_per_bin=15,
-        verbose=False,
     ):
         key = cache_key(
             manifest,
@@ -123,40 +139,69 @@ class Annotation:
         cls._cache[key] = instance
         return instance
 
+    def __getnewargs__(self):
+        # Necessary for pickle
+        return (
+            self._init_manifest,
+            self._init_array_type,
+            self._init_gap,
+            self._init_detail,
+            self._init_bin_size,
+            self._init_min_probes_per_bin,
+        )
+
     def __init__(
         self,
         manifest=None,
         array_type=None,
-        gap=UNSET,
-        detail=UNSET,
+        gap=None,
+        detail=None,
         bin_size=50000,
         min_probes_per_bin=15,
-        verbose=False,
     ):
         # Don't need to initialize if instance is cached.
         if hasattr(self, "_cached"):
             return
         self._cached = True
 
+        self._init_manifest = manifest
+        self._init_array_type = array_type
+        self._init_gap = gap
+        self._init_detail = detail
+        self._init_bin_size = bin_size
+        self._init_min_probes_per_bin = min_probes_per_bin
+
+        pickle_hash = input_args_id(
+            "annotation",
+            self._init_manifest,
+            self._init_array_type,
+            self._init_gap,
+            self._init_detail,
+            self._init_bin_size,
+            self._init_min_probes_per_bin,
+        )
+        self._pickle_path = MEPYLOME_TMP_DIR / f"{pickle_hash}.pkl"
+        if self._pickle_path.exists():
+            with self._pickle_path.open("rb") as file:
+                saved_instance = pickle.load(file)
+                self.__dict__.update(saved_instance.__dict__)
+                return
+
+        logger.info("Constructing annotation...")
         if manifest is None and array_type is None:
             msg = "'manifest' or 'array_type' must be given"
             raise ValueError(msg)
         self.bin_size = bin_size
         self.min_probes_per_bin = min_probes_per_bin
 
-        self.gap = gap
-        if self.gap is UNSET:
-            self.gap = self.default_gaps()
+        self.gap = gap or self.default_gaps()
 
-        self.detail = detail
-        if self.detail is UNSET:
-            self.detail = self.default_genes()
-        elif self.detail is not None:
+        self.detail = detail or self.default_genes()
+        if detail is not None:
             # PyRanges ranges start at 0
             self.detail.Start -= 1
             self.detail = self.detail.sort()
 
-        self.verbose = verbose
         self.chromsizes = pr.data.chromsizes()
 
         if array_type is None:
@@ -197,6 +242,12 @@ class Annotation:
                 self.adjusted_manifest[["IlmnID"]]
             ).df[["Name", "IlmnID"]]
 
+        # Save to disk
+        with self._pickle_path.open("wb") as file:
+            pickle.dump(self, file)
+
+        logger.info("Constructing annotation done")
+
     @staticmethod
     @lru_cache
     def default_gaps():
@@ -238,13 +289,13 @@ class Annotation:
         )
         return bins.apply(
             lambda df: self.merge_bins_in_chromosome(
-                df, self.min_probes_per_bin, verbose=self.verbose
+                df, self.min_probes_per_bin
             ),
             as_pyranges=True,
         )
 
     @staticmethod
-    def merge_bins_in_chromosome(bin_df, min_probes_per_bin, verbose=False):
+    def merge_bins_in_chromosome(bin_df, min_probes_per_bin):
         """Merges adjacent bins until all contain a minimum of probes.
 
         Args:
@@ -252,8 +303,6 @@ class Annotation:
                 single chromosome.
             min_probes_per_bin (int): Minimum number of probes per bin required
                 for merging.
-            verbose (bool, optional): Whether to print verbose output. Defaults
-                to False.
 
         Returns:
             DataFrame: Merged bins in the chromosome.
@@ -318,16 +367,15 @@ class Annotation:
             # Invalid
             if n_probes_left == INVALID and n_probes_right == INVALID:
                 update_neighbors(i_left, i_min, i_right)
-                if verbose:
+                if logger.isEnabledFor(logging.DEBUG):
                     row = (
                         bin_df.loc[i_min, ["Chromosome", "Start", "End"]]
                         .astype(str)
                         .tolist()
                     )
                     row_str = "-".join(row)
-                    log(
-                        f"[Annotation] Could not merge {row_str}. "
-                        "Removed instead."
+                    logger.debug(
+                        "Could not merge %s. Removed instead.", row_str
                     )
                 continue
             if n_probes_right == INVALID or n_probes_left <= n_probes_right:
@@ -349,6 +397,36 @@ class Annotation:
         bin_df = bin_df[bin_df.N_probes != INVALID]
         return bin_df[["Chromosome", "Start", "End", "N_probes"]]
 
+    @staticmethod
+    def load(array_types=None):
+        """Loads specified annotation into memory.
+
+        Args:
+            array_types (list or ArrayType, optional): List of array types or
+                a single array type to load. Defaults to all available types.
+
+        Examples:
+            >>> # Load all annotations:
+            >>> Annotation.load()
+
+            >>> # Load specific annotation:
+            >>> Annotation.load(
+            >>>     [ArrayType.ILLUMINA_450K, ArrayType.ILLUMINA_EPIC]
+            >>> )
+            >>> Annotation.load("epicv2")
+        """
+        if array_types is None:
+            array_types = [
+                ArrayType.ILLUMINA_450K,
+                ArrayType.ILLUMINA_EPIC,
+                ArrayType.ILLUMINA_EPIC_V2,
+                ArrayType.ILLUMINA_MSA48,
+            ]
+        if not isinstance(array_types, list):
+            array_types = [array_types]
+        for array_type in array_types:
+            _ = Annotation(array_type=array_type)
+
     def __repr__(self):
         title = "Annotation():"
         lines = [
@@ -390,7 +468,6 @@ class CNV:
             neutral references.
         annotation (Annotation): Annotation object containing genomic
             annotation information.
-        verbose (bool): Whether to print verbose output.
         bins (PyRanges): PyRanges object representing genomic bins.
         probes (Index): Index of probe IDs.
         coef: Coefficient of linear regression.
@@ -410,8 +487,6 @@ class CNV:
         annotation (Annotation, optional): Annotation object containing
             genomic annotation information. Defaults to annotation
             associated with the sample array type.
-        verbose (bool, optional): Whether to print verbose output. Defaults
-            to False.
 
     Examples:
         >>> sample = MethylData(file="path/to/idat/file")
@@ -435,7 +510,7 @@ class CNV:
         https://doi.org/10.1093/bioinformatics/btae029
     """
 
-    def __init__(self, sample, reference, annotation=None, *, verbose=False):
+    def __init__(self, sample, reference, annotation=None):
         if len(sample.probes) != 1:
             msg = "sample must contain exactly 1 probe."
             raise ValueError(msg)
@@ -451,10 +526,7 @@ class CNV:
                 "or 'ReferenceMethylData'"
             )
             raise TypeError(msg)
-        self.verbose = verbose
         self.annotation = annotation
-        if self.verbose:
-            log(f"[CNV] {self.probe} Construct annotation...")
         if annotation is None:
             self.annotation = Annotation(
                 array_type=sample.array_type,
@@ -485,9 +557,7 @@ class CNV:
         self.fit()
 
     @classmethod
-    def set_all(
-        cls, sample, reference, annotation=None, *, do_seg=True, verbose=False
-    ):
+    def set_all(cls, sample, reference, annotation=None, *, do_seg=True):
         """Create a CNV object and perform CNV analysis.
 
         Args:
@@ -519,7 +589,6 @@ class CNV:
             sample=sample,
             reference=reference,
             annotation=annotation,
-            verbose=verbose,
         )
         cnv.set_bins()
         cnv.set_detail()
@@ -531,8 +600,7 @@ class CNV:
         """Calculates intensity values from methylation data."""
         if hasattr(methyl_data, "intensity"):
             return
-        if self.verbose:
-            log(f"[CNV] {self.probe} Setting intensity...")
+        logger.debug("%s Setting intensity...", self.probe)
         intensity = methyl_data.methyl + methyl_data.unmethyl
         prefix = (
             f"{self.probe}"
@@ -544,22 +612,22 @@ class CNV:
         nan_indices = np.isnan(intensity)
         if np.any(nan_indices):
             intensity[nan_indices] = 1
-            if self.verbose:
-                log(f"[CNV] {prefix}: Intensities that are NA set to 1.")
+            logger.debug("%s: Intensities that are NA set to 1", prefix)
 
         # Replace values less than 1 with 1
         lt_one_indices = intensity < 1
         if np.any(lt_one_indices):
             intensity[lt_one_indices] = 1
-            if self.verbose:
-                log(f"[CNV] {prefix}: Intensities < 0 set to 1.")
+            logger.debug("%s: Intensities < 0 set to 1", prefix)
 
         # Check abnormal low and high intensities
         mean_intensity = np.mean(intensity, axis=1)
-        if np.min(mean_intensity) < 5000 and self.verbose:
-            log(f"[CNV] {prefix}: Intensities are abnormally low (< 5000).")
-        if np.max(mean_intensity) > 50000 and self.verbose:
-            log(f"[CNV] {prefix}: Intensities are abnormally high (> 50000).")
+        if np.min(mean_intensity) < 5000:
+            logger.info("%s: Intensities are abnormally low (< 5000)", prefix)
+        if np.max(mean_intensity) > 50000:
+            logger.info(
+                "%s: Intensities are abnormally high (> 50000)", prefix
+            )
         methyl_data.intensity = pd.DataFrame(
             intensity.T,
             columns=methyl_data.probes,
@@ -572,8 +640,7 @@ class CNV:
         This method fits a linear regression model to the intensity data of the
         sample and reference and calculates the CNV at every CpG site.
         """
-        if self.verbose:
-            log(f"[CNV] {self.probe} Performing fit...")
+        logger.info("%s Performing fit...", self.probe)
         from sklearn.linear_model import LinearRegression
 
         smp_intensity = _pd_loc(
@@ -585,11 +652,10 @@ class CNV:
             [np.corrcoef(smp_intensity, z)[0, 1] for z in ref_intensity.T]
         )
         if any(correlation >= 0.99):
-            if self.verbose:
-                log(
-                    f"[CNV] {self.probe} Sample found in reference set. "
-                    "Excluded from fitting."
-                )
+            logger.info(
+                "%s Sample found in reference set. Excluded from fitting.",
+                self.probe,
+            )
             ref_intensity = ref_intensity[:, correlation < 0.99]
         X = np.log2(ref_intensity)
         y = np.log2(smp_intensity)
@@ -610,8 +676,7 @@ class CNV:
         by taking the median of the ratios obtained from the linear regression
         model fit in the 'fit' method.
         """
-        if self.verbose:
-            log(f"[CNV] {self.probe} Setting bins...")
+        logger.info("%s Setting bins...", self.probe)
         cpg_bins = self.annotation._cpg_bins.copy()
         cpg_bins["ratio"] = _pd_loc(self.ratio, cpg_bins.IlmnID).ratio.values
         result = cpg_bins.groupby("bins_index", dropna=False)["ratio"].agg(
@@ -632,8 +697,7 @@ class CNV:
         specified in the detail object. The result includes the median ratio,
         variance, and count of probes within each region.
         """
-        if self.verbose:
-            log(f"[CNV] {self.probe} Setting detail...")
+        logger.info("%s Setting detail...", self.probe)
         cpg_detail = self.annotation._cpg_detail.copy()
         cpg_detail["ratio"] = _pd_loc(
             self.ratio, cpg_detail.IlmnID
@@ -677,10 +741,7 @@ class CNV:
         chrom = df["Chromosome"].iloc[0]
         seg = cbsegment(bin_values)
         return pd.DataFrame(
-            [
-                [chrom, df.Start.iloc[s.start], df.End.iloc[s.end - 1]]
-                for s in seg
-            ],
+            [[chrom, df.Start.iloc[s[0]], df.End.iloc[s[1] - 1]] for s in seg],
             columns=["Chromosome", "Start", "End"],
         )
 
@@ -692,10 +753,9 @@ class CNV:
         It calculates the CNV segments for each chromosome and stores them
         in the 'segments' attribute of the object.
         """
-        if _get_cgsegment(verbose=True) is None:
+        if _get_cgsegment() is None:
             return
-        if self.verbose:
-            log(f"[CNV] {self.probe} Setting segments...")
+        logger.info("%s Setting segments...", self.probe)
         segments = self.bins.apply(self._get_segments)
         overlap = segments.join(self.annotation.adjusted_manifest[["IlmnID"]])
         overlap.ratio = self.ratio.loc[overlap.IlmnID].ratio
@@ -733,8 +793,7 @@ class CNV:
         Raises:
             ValueError: If an invalid data option is specified.
         """
-        if self.verbose:
-            log(f"[CNV] {self.probe} Write data to disk...")
+        logger.info("%s Write data to disk...", self.probe)
         default = {"all", "bins", "detail", "segments", "metadata"}
         if not isinstance(data, list):
             data = [data]
@@ -778,8 +837,7 @@ class CNV:
 
     def plot(self):
         """Generates and displays a plot of the CNV data."""
-        if self.verbose:
-            log(f"[CNV] {self.probe} Plotting...")
+        logger.info("%s Plotting...", self.probe)
         cnv_dir = Path(MEPYLOME_TMP_DIR, "cnv_zips")
         ensure_directory_exists(cnv_dir)
         cnv_file = self.probe + ZIP_ENDING
