@@ -5,6 +5,7 @@ Illumina array manifest files, which contain information about probes and their
 characteristics.
 """
 
+import logging
 import pickle
 from pathlib import Path
 
@@ -17,13 +18,14 @@ from mepylome.dtypes.cache import cache_key, input_args_id
 from mepylome.dtypes.chromosome import Chromosome
 from mepylome.dtypes.probes import Channel, InfiniumDesignType, ProbeType
 from mepylome.utils.files import (
-    MEPYLOME_TMP_DIR,
     download_file,
     ensure_directory_exists,
     get_csv_file,
     reset_file,
 )
-from mepylome.utils.varia import log
+from mepylome.utils.varia import CONFIG, MEPYLOME_TMP_DIR
+
+logger = logging.getLogger(__name__)
 
 __all__ = ["Manifest"]
 
@@ -31,37 +33,20 @@ __all__ = ["Manifest"]
 MANIFEST_DIR = Path.home() / ".mepylome" / "manifest_files_v0"
 DOWNLOAD_DIR = MEPYLOME_TMP_DIR / "manifests"
 
-ENDING_CONTROL_PROBES = "_control-probes"
+ENDING_CONTROL_PROBES = CONFIG["suffixes"]["manifest_control_probes"]
 NONE = -1
 
 MANIFEST_URL = {
-    ArrayType.ILLUMINA_450K: (
-        "https://webdata.illumina.com/downloads/productfiles/humanmethylation450/humanmethylation450_15017482_v1-2.csv"
-    ),
-    ArrayType.ILLUMINA_EPIC: (
-        "https://webdata.illumina.com/downloads/productfiles/methylationEPIC/infinium-methylationepic-v-1-0-b5-manifest-file-csv.zip"
-    ),
-    ArrayType.ILLUMINA_EPIC_V2: (
-        "https://support.illumina.com.cn/content/dam/illumina-support/documents/downloads/productfiles/methylationepic/InfiniumMethylationEPICv2.0ProductFiles(ZIPFormat).zip"
-    ),
-    ArrayType.ILLUMINA_MOUSE: (
-        "https://support.illumina.com/content/dam/illumina-support/documents/downloads/productfiles/mouse-methylation/infinium-mouse-methylation-manifest-file-csv.zip"
-    ),
+    ArrayType(type_): url for type_, url in CONFIG["urls"]["manifest"].items()
 }
-PROCESSED_MANIFEST_URL = f"https://raw.githubusercontent.com/brj0/mepylome-data/main/{MANIFEST_DIR.name}/"
-
+PROCESSED_MANIFEST_URL = (
+    CONFIG["urls"]["processed_manifests"] % MANIFEST_DIR.name
+)
 REMOTE_FILENAME = {
-    ArrayType.ILLUMINA_450K: "humanmethylation450_15017482_v1-2.csv",
-    ArrayType.ILLUMINA_EPIC: "infinium-methylationepic-v-1-0-b5-manifest-file.csv",
-    ArrayType.ILLUMINA_EPIC_V2: "EPIC-8v2-0_A1.csv",
-    ArrayType.ILLUMINA_MOUSE: "MouseMethylation-12v1-0_A2.csv",
+    ArrayType(type_): url for type_, url in CONFIG["files"]["remote"].items()
 }
-
 LOCAL_FILENAME = {
-    ArrayType.ILLUMINA_450K: "manifest-450k.csv.gz",
-    ArrayType.ILLUMINA_EPIC: "manifest-epic.csv.gz",
-    ArrayType.ILLUMINA_EPIC_V2: "manifest-epicv2.csv.gz",
-    ArrayType.ILLUMINA_MOUSE: "manifest-mouse.csv.gz",
+    ArrayType(type_): url for type_, url in CONFIG["files"]["local"].items()
 }
 
 
@@ -73,6 +58,7 @@ PROBES_COLUMNS = [
     "Infinium_Design_Type",
     "Color_Channel",
     "CHR",
+    "Chr",
     "MAPINFO",
     "AlleleA_ProbeSeq",
     "AlleleB_ProbeSeq",
@@ -102,7 +88,7 @@ class Manifest:
         array_type (str or ArrayType): The type of array to process. Use either
             ArrayType (ArrayType.ILLUMINA_450K, ArrayType.ILLUMINA_EPIC,
             ArrayType.ILLUMINA_EPIC_V2) or corresponding string ('450k',
-            'epic', 'epicv2')
+            'epic', 'epicv2', 'msa48')
         proc_path (str or Path): The path to the local processed manifest file
             (default: None).
         raw_path (str or Path, optional): Path to the raw manifest file.
@@ -146,10 +132,10 @@ class Manifest:
     def __getnewargs__(self):
         # Necessary for pickle
         return (
-            self.array_type,
-            self.raw_path,
-            self.proc_path,
-            self.download_proc,
+            self._init_array_type,
+            self._init_raw_path,
+            self._init_proc_path,
+            self._init_download_proc,
         )
 
     def __init__(
@@ -163,6 +149,11 @@ class Manifest:
             return
         self._cached = True
 
+        self._init_array_type = array_type
+        self._init_raw_path = raw_path
+        self._init_proc_path = proc_path
+        self._init_download_proc = download_proc
+
         def to_path(x):
             return x if x is None else Path(x)
 
@@ -183,10 +174,10 @@ class Manifest:
                 return
 
         if self.array_type == ArrayType.UNKNOWN:
-            self.__data_frame = pd.DataFrame()
-            self.__control_data_frame = pd.DataFrame()
-            self.__snp_data_frame = pd.DataFrame()
-            self.__methyl_probes = None
+            self._data_frame = pd.DataFrame()
+            self._control_data_frame = pd.DataFrame()
+            self._snp_data_frame = pd.DataFrame()
+            self._methyl_probes = None
             return
 
         # Set processed manifest path
@@ -227,10 +218,10 @@ class Manifest:
                 msg = "Provide either array_type or proc_path or raw_path"
                 raise ValueError(msg)
 
-        self.__data_frame = self._read_probes(self.proc_path)
-        self.__control_data_frame = self._read_control_probes(self.ctrl_path)
-        self.__snp_data_frame = self._read_snp_probes()
-        self.__methyl_probes = None
+        self._data_frame = self._read_probes(self.proc_path)
+        self._control_data_frame = self._read_control_probes(self.ctrl_path)
+        self._snp_data_frame = self._read_snp_probes()
+        self._methyl_probes = None
 
         # Save to disk
         with self._pickle_path.open("wb") as file:
@@ -239,22 +230,22 @@ class Manifest:
     @property
     def data_frame(self):
         """Pandas data frame of all manifest probes."""
-        return self.__data_frame
+        return self._data_frame
 
     @property
     def control_data_frame(self):
         """Pandas data frame of all manifest control probes."""
-        return self.__control_data_frame
+        return self._control_data_frame
 
     @property
     def snp_data_frame(self):
         """SNP probes from the manifest data frame."""
-        return self.__snp_data_frame
+        return self._snp_data_frame
 
     @property
     def methylation_probes(self):
         """All type I and II probes."""
-        if self.__methyl_probes is None:
+        if self._methyl_probes is None:
             type_1 = self.probe_info(ProbeType.ONE)
             type_2 = self.probe_info(ProbeType.TWO)
             idx = np.sort(
@@ -265,19 +256,19 @@ class Manifest:
                     ]
                 )
             )
-            self.__methyl_probes = self.__data_frame.iloc[idx]["IlmnID"].values
-        return self.__methyl_probes
+            self._methyl_probes = self._data_frame.iloc[idx]["IlmnID"].values
+        return self._methyl_probes
 
     def control_address(self, control_type=None):
         """Returns address IDs of all control probes of the specified type."""
         if control_type is None:
-            return self.__control_data_frame.Address_ID
+            return self._control_data_frame.Address_ID
         # Ensure control_type is a list-like object
         if not isinstance(control_type, (list, tuple)):
             control_type = [control_type]
         # Use isin() with the list-like object
-        return self.__control_data_frame[
-            self.__control_data_frame.Control_Type.isin(control_type)
+        return self._control_data_frame[
+            self._control_data_frame.Control_Type.isin(control_type)
         ].Address_ID
 
     @staticmethod
@@ -303,6 +294,7 @@ class Manifest:
                 ArrayType.ILLUMINA_450K,
                 ArrayType.ILLUMINA_EPIC,
                 ArrayType.ILLUMINA_EPIC_V2,
+                ArrayType.ILLUMINA_MSA48,
             ]
         if not isinstance(array_types, list):
             array_types = [array_types]
@@ -324,17 +316,17 @@ class Manifest:
         source_url = MANIFEST_URL[self.array_type]
         source_filename = Path(source_url).name
 
-        log(f"[Manifest] Downloading manifest: {source_filename}")
+        logger.info("Downloading manifest: %s", source_filename)
 
         self.raw_path = DOWNLOAD_DIR / source_filename
         download_file(source_url, self.raw_path)
 
     def _download_processed_manifest(self):
         """Download processed manifest file and return true if successful."""
-        log(f"[Manifest] Downloading processed {self.array_type} manifest")
+        logger.info("Downloading processed %s manifest", self.array_type)
         try:
             for path in [self.proc_path, self.ctrl_path]:
-                ensure_directory_exists(path)
+                ensure_directory_exists(path.parent)
                 url = PROCESSED_MANIFEST_URL + path.name
                 download_file(url, path)
             return True
@@ -360,18 +352,23 @@ class Manifest:
                 archive. If not provided, it defaults to the name of the
                 raw_path file.
         """
-        log(f"[Manifest] Process raw manifest {self.raw_path}")
+        logger.info("Process raw manifest %s", self.raw_path)
         if csv_filename is None:
             csv_filename = self.raw_path.name
-        ensure_directory_exists(self.proc_path)
-        ensure_directory_exists(self.ctrl_path)
+        ensure_directory_exists(self.proc_path.parent)
+        ensure_directory_exists(self.ctrl_path.parent)
         with get_csv_file(self.raw_path, csv_filename) as manifest_file:
             # Process probes
             Manifest._seek_to_start(manifest_file)
+            available_columns = pd.read_csv(manifest_file, nrows=0).columns
+            Manifest._seek_to_start(manifest_file)
+            valid_columns = [
+                col for col in PROBES_COLUMNS if col in available_columns
+            ]
             probes_df = pd.read_csv(
                 manifest_file,
                 low_memory=False,
-                usecols=PROBES_COLUMNS,
+                usecols=valid_columns,
             )
             n_probes = probes_df[probes_df.IlmnID.str.startswith("[")].index[0]
             probes_df = probes_df[:n_probes]
@@ -387,62 +384,81 @@ class Manifest:
                 usecols=range(len(CONTROL_COLUMNS)),
             )
             controls_df.columns = CONTROL_COLUMNS
-            controls_df["Address_ID"] = controls_df["Address_ID"].astype(
-                "int32"
-            )
+            if (
+                pd.to_numeric(controls_df["Address_ID"], errors="coerce")
+                .notna()
+                .all()
+            ):
+                controls_df["Address_ID"] = controls_df["Address_ID"].astype(
+                    "int32"
+                )
             controls_df.to_csv(self.ctrl_path, index=False)
 
     @staticmethod
     def _process_probes(data_frame):
         """Transforms manifest probes to a more efficient internal format."""
-        data_frame = data_frame.rename(
-            columns={
-                "MAPINFO": "Start",
-                "CHR": "Chromosome",
-            },
-        )
+        rename_map = {
+            "MAPINFO": "Start",
+            "CHR": "Chromosome",
+            "Chr": "Chromosome",
+        }
+        data_frame = data_frame.rename(columns=rename_map)
         data_frame["Chromosome"] = Chromosome.pd_from_string(
             data_frame["Chromosome"]
         )
         # IlmnID and Name are different in EPICv2
-        data_frame.IlmnID = data_frame.Name
-        data_frame = data_frame.drop(columns="Name")
+        data_frame["IlmnID"] = data_frame["Name"]
+        data_frame = data_frame.drop(columns=["Name"])
         channel_to_int = {"Grn": Channel.GRN, "Red": Channel.RED}
+        design_type_map = {
+            "I": InfiniumDesignType.I,
+            "II": InfiniumDesignType.II,
+        }
         with pd.option_context("future.no_silent_downcasting", True):
-            data_frame["Color_Channel"] = (
-                data_frame["Color_Channel"]
-                .replace(channel_to_int)
-                .infer_objects()
-            )
-            data_frame["Infinium_Design_Type"] = (
-                data_frame["Infinium_Design_Type"]
-                .replace(
-                    {"I": InfiniumDesignType.I, "II": InfiniumDesignType.II}
+            if "Color_Channel" in data_frame.columns:
+                data_frame["Color_Channel"] = (
+                    data_frame["Color_Channel"]
+                    .replace(channel_to_int)
+                    .infer_objects()
                 )
-                .infer_objects()
+
+            if "Infinium_Design_Type" in data_frame.columns:
+                data_frame["Infinium_Design_Type"] = (
+                    data_frame["Infinium_Design_Type"]
+                    .replace(design_type_map)
+                    .infer_objects()
+                )
+
+        data_frame["TypeI_N_CpG"] = 0
+        data_frame["TypeII_N_CpG"] = 0
+
+        if "AlleleB_ProbeSeq" in data_frame.columns:
+            data_frame["TypeI_N_CpG"] = np.maximum(
+                0,
+                data_frame["AlleleB_ProbeSeq"].fillna("").str.count("CG") - 1,
             )
-        data_frame["TypeI_N_CpG"] = np.maximum(
-            0, data_frame["AlleleB_ProbeSeq"].fillna("").str.count("CG") - 1
-        )
-        # R Stands for CG in AlleleA_ProbeSeq
-        data_frame["TypeII_N_CpG"] = data_frame["AlleleA_ProbeSeq"].str.count(
-            "R"
-        )
+
+        if "AlleleA_ProbeSeq" in data_frame.columns:
+            # R Stands for CG in AlleleA_ProbeSeq
+            data_frame["TypeII_N_CpG"] = (
+                data_frame["AlleleA_ProbeSeq"].fillna("").str.count("R")
+            )
+
         data_frame["N_CpG"] = NONE
-        data_frame.loc[
-            data_frame["Infinium_Design_Type"] == InfiniumDesignType.I,
-            "N_CpG",
-        ] = data_frame.loc[
-            data_frame["Infinium_Design_Type"] == InfiniumDesignType.I,
-            "TypeI_N_CpG",
-        ]
-        data_frame.loc[
-            data_frame["Infinium_Design_Type"] == InfiniumDesignType.II,
-            "N_CpG",
-        ] = data_frame.loc[
-            data_frame["Infinium_Design_Type"] == InfiniumDesignType.II,
-            "TypeII_N_CpG",
-        ]
+        if "Infinium_Design_Type" in data_frame.columns:
+            is_type_I = (
+                data_frame["Infinium_Design_Type"] == InfiniumDesignType.I
+            )
+            is_type_II = (
+                data_frame["Infinium_Design_Type"] == InfiniumDesignType.II
+            )
+
+            data_frame.loc[is_type_I, "N_CpG"] = data_frame.loc[
+                is_type_I, "TypeI_N_CpG"
+            ]
+            data_frame.loc[is_type_II, "N_CpG"] = data_frame.loc[
+                is_type_II, "TypeII_N_CpG"
+            ]
 
         # Use int32 to improve performance of indexing
         int_cols = [
@@ -454,25 +470,38 @@ class Manifest:
             "TypeI_N_CpG",
             "TypeII_N_CpG",
         ]
-        data_frame[int_cols] = (
-            data_frame[int_cols].fillna(NONE).astype("int32")
-        )
-        data_frame["End"] = data_frame["Start"]
-        data_frame = data_frame.drop(
-            columns=["AlleleA_ProbeSeq", "AlleleB_ProbeSeq"]
-        )
+        for col in int_cols:
+            if col in data_frame.columns:
+                data_frame[col] = data_frame[col].fillna(NONE).astype("int32")
+
+        if "Start" in data_frame.columns:
+            data_frame["End"] = data_frame["Start"]
+
+        drop_cols = ["AlleleA_ProbeSeq", "AlleleB_ProbeSeq"]
+        existing_drop_cols = [
+            col for col in drop_cols if col in data_frame.columns
+        ]
+        data_frame = data_frame.drop(columns=existing_drop_cols)
 
         def get_probe_type(name, infinium_type):
             """Determines the probe type (I, II, SnpI, SnpII or Control)."""
             probe_type = ProbeType.from_manifest_values(name, infinium_type)
             return probe_type.value
 
-        vectorized_get_type = np.vectorize(get_probe_type)
-        data_frame["Probe_Type"] = vectorized_get_type(
-            data_frame.IlmnID.values,
-            data_frame["Infinium_Design_Type"].values,
-        )
-        # TODO drop Infinium_Design_Type
+        if (
+            "IlmnID" in data_frame.columns
+            and "Infinium_Design_Type" in data_frame.columns
+        ):
+            vectorized_get_type = np.vectorize(get_probe_type)
+            data_frame["Probe_Type"] = vectorized_get_type(
+                data_frame["IlmnID"].values,
+                data_frame["Infinium_Design_Type"].values,
+            )
+
+        # TODO: drop Infinium_Design_Type
+        if not {"Chromosome", "Start", "End"}.issubset(data_frame.columns):
+            return data_frame
+
         probes_ranges = pr.PyRanges(data_frame).sort()
         return probes_ranges.df
 
@@ -510,7 +539,7 @@ class Manifest:
             probes_file,
             dtype=self._get_data_types(),
         )
-        # TODO epicv2 has duplicated ID's for example c='cg22367159'
+        # TODO: epicv2 has duplicated ID's for example c='cg22367159'
         data_frame = data_frame.drop_duplicates(
             subset=["IlmnID"], keep="first"
         )
@@ -523,9 +552,14 @@ class Manifest:
             dtype=self._get_data_types(),
         )
         # Use int32 instead of int64 to improve performance of indexing
-        data_frame["Address_ID"] = (
-            data_frame["Address_ID"].astype("int32").copy()
-        )
+        if (
+            pd.to_numeric(data_frame["Address_ID"], errors="coerce")
+            .notna()
+            .all()
+        ):
+            data_frame["Address_ID"] = (
+                data_frame["Address_ID"].astype("int32").copy()
+            )
         return data_frame
 
     def _read_snp_probes(self):
