@@ -1,6 +1,7 @@
 """Contains utilities to generate CNV plots."""
 
 import io
+import logging
 import threading
 import webbrowser
 import zipfile
@@ -9,20 +10,22 @@ from pathlib import Path
 
 import numpy as np
 import pandas as pd
+import plotly
 import plotly.graph_objects as go
 
 from mepylome.dtypes.cache import memoize
-from mepylome.dtypes.genetic_data import CHROMOSOME_DATA, IMPORTANT_GENES
-from mepylome.utils import log
-from mepylome.utils.files import MEPYLOME_TMP_DIR, ensure_directory_exists
-from mepylome.utils.varia import get_free_port
+from mepylome.utils.files import ensure_directory_exists
+from mepylome.utils.varia import CONFIG, MEPYLOME_TMP_DIR, get_free_port
+
+logger = logging.getLogger(__name__)
 
 PLOTLY_RENDER_MODE = "webgl"
-ZIP_ENDING = "_cnv.zip"
+ZIP_ENDING = CONFIG["suffixes"]["cnv_zip"]
 THRESHOLD_BALANCED = 0.1
+CHROMOSOME_DATA = CONFIG["genes"]["chromosome_data"]
 
 ensure_directory_exists(MEPYLOME_TMP_DIR)
-CNV_GRID = Path(MEPYLOME_TMP_DIR, "cnv_grid.json")
+CNV_GRID = Path(MEPYLOME_TMP_DIR, f"cnv_grid_v{plotly.__version__}.json")
 
 
 class Genome:
@@ -68,8 +71,7 @@ offset = {i.name: i.offset for i in reference_genome}
 
 def get_df_from_zip(zip_file_path, extract=None):
     """Reads the CNV data from a zip file without extraxtion on disk."""
-    if extract is None:
-        extract = ["bins", "detail", "segments"]
+    extract = extract or ["bins", "detail", "segments"]
     zip_file_path = Path(zip_file_path).expanduser()
     zip_file = zip_file_path.name
     csv_file_templ = zip_file.replace(ZIP_ENDING, "_cnv_%s.csv")
@@ -154,9 +156,12 @@ def cnv_bins_plot(data_frame, title, labels):
         x=data_frame["x"],
         y=data_frame["y"],
         mode="markers",
-        marker=dict(
-            color=data_frame["y"], colorscale="RdBu_r", cmin=-0.4, cmax=0.4
-        ),
+        marker={
+            "color": data_frame["y"],
+            "colorscale": "RdBu_r",
+            "cmin": -0.4,
+            "cmax": 0.4,
+        },
         text=data_frame["hover_data"],
         hovertemplate=(
             "<b>Value:</b> %{y}<br><b>Genes:</b> %{text}<extra></extra>"
@@ -272,16 +277,21 @@ def find_genes_within_bins(bins, detail):
     """
     bins_X_start = add_offset(bins, "Chromosome", "Start")
     bins_X_end = add_offset(bins, "Chromosome", "End")
+
     if not bins_X_start.is_monotonic_increasing:
         msg = "'bins' is not sorted"
         raise ValueError(msg)
+
     detail_X_start = add_offset(detail, "Chromosome", "Start")
     detail_X_end = add_offset(detail, "Chromosome", "End")
+
     detail_Start_bin = np.digitize(detail_X_start, bins_X_start) - 1
     detail_Start_bin_ = np.digitize(detail_X_start, bins_X_end, right=True)
     detail_End_bin = np.digitize(detail_X_end, bins_X_start) - 1
+
     # Identify genes that start within gaps between bins
     detail_Start_bin[detail_Start_bin < detail_Start_bin_] += 1
+
     detail_Range = [
         np.arange(start, end)
         for start, end in zip(detail_Start_bin, detail_End_bin + 1)
@@ -298,7 +308,7 @@ def find_genes_within_bins(bins, detail):
         # By sorting first, resultig string will be sorted too
         .sort_values(by="Name")
         .groupby("Range")["Name"]
-        .agg(lambda x: ", ".join(x))
+        .agg(", ".join)
         .reindex(range(len(bins)), fill_value="")
         .reset_index(drop=True)
     )
@@ -332,16 +342,17 @@ def _find_genes_within_bins(bins, detail):
 
 def read_cnv_data_from_disk(cnv_dir, sample_id, extract=None):
     """Reads the components of a zip-file generated by CNV.write."""
-    if extract is None:
-        extract = ["bins", "detail", "segments"]
+    extract = extract or ["bins", "detail", "segments"]
     if isinstance(extract, str):
         extract = [extract]
+
     sample_zip = sample_id + ZIP_ENDING
     unzipped = get_df_from_zip(
-        cnv_dir / sample_zip,
+        Path(cnv_dir) / sample_zip,
         extract=extract,
     )
-    results = {key: value for key, value in zip(extract, unzipped)}
+
+    results = dict(zip(extract, unzipped))
     # Calculate some plot x-values
     if results.get("bins") is not None:
         results["bins"]["X_mid"] = get_x_mid(results["bins"])
@@ -357,14 +368,14 @@ def read_cnv_data_from_disk(cnv_dir, sample_id, extract=None):
         results["segments"]["X_end"] = add_offset(
             results["segments"], "Chromosome", "End"
         )
+
     if len(results) == 1:
         return next(iter(results.values()))
-    else:
-        return tuple(results.values())
+    return tuple(results.values())
 
 
 def cnv_plot_from_data(
-    sample_id, bins, detail, segments, genes_fix, genes_sel, *, verbose=False
+    sample_id, bins, detail, segments, genes_fix, genes_sel
 ):
     """Generate a CNV plot from data calculated by the class CNV.
 
@@ -380,8 +391,7 @@ def cnv_plot_from_data(
     Returns:
         Plotly Figure: A Plotly figure representing the CNV plot.
     """
-    if verbose:
-        log("[CNV-Plot] Make CNV plot: prepare data...")
+    logger.info("Make CNV plot: prepare data...")
     bins_Genes, detail_Range = find_genes_within_bins(
         bins[["Chromosome", "Start", "End"]],
         detail[["Chromosome", "Start", "End", "Name"]],
@@ -392,8 +402,7 @@ def cnv_plot_from_data(
     scatter_df.columns = ["x", "y", "hover_data"]
 
     # Base scatterplot
-    if verbose:
-        log("[CNV-Plot] Make CNV plot: bins...")
+    logger.info("Make CNV plot: bins...")
     plot = cnv_bins_plot(
         data_frame=scatter_df,
         title=f"Sample ID: {sample_id}",
@@ -401,12 +410,25 @@ def cnv_plot_from_data(
     )
 
     # Highlight bins adjacent to the added genes
-    if verbose:
-        log("[CNV-Plot] Make CNV plot: genes...")
+    logger.info("Make CNV plot: genes...")
     genes_sel = genes_sel if genes_sel else []
-    genes_x_range = (
-        detail[detail["Name"].isin(genes_sel)]["Range"].explode().tolist()
-    )
+
+    selected_genes_df = detail[detail["Name"].isin(genes_sel)]
+
+    empty_range_genes = selected_genes_df["Name"][
+        selected_genes_df["Range"].apply(lambda x: len(x) == 0)
+    ].tolist()
+
+    if empty_range_genes:
+        msg = (
+            "The following selected genes are outside of the CNV bins "
+            "and therefore cannot be highlighted: "
+            + ", ".join(empty_range_genes)
+        )
+        raise ValueError(msg)
+
+    genes_x_range = selected_genes_df["Range"].explode().tolist()
+
     highlighted_bins = bins.loc[genes_x_range, ["X_mid", "Median"]]
     plot = add_highlited_bins(plot, highlighted_bins)
 
@@ -416,8 +438,7 @@ def cnv_plot_from_data(
     plot = add_genes(plot, gene_detail)
 
     # Draw the segments
-    if verbose:
-        log("[CNV-Plot] Make CNV plot: segments...")
+    logger.info("Make CNV plot: segments...")
     plot = add_segments(plot, segments)
     return plot
 
@@ -432,7 +453,7 @@ class CNVPlot:
         cnv_dir (str): Path to the directory containing CNV data.
         cnv_file (str): Name of the CNV file in the directory.
         genes_fix (list): List of genes to highlight. Defaults to
-            IMPORTANT_GENES.
+            CONFIG["genes"]["default_genes_list"].
         host (str): Host address for running the Dash app. Defaults to
             "localhost".
         port (int): Port number for running the Dash app. Defaults to 8050.
@@ -443,7 +464,7 @@ class CNVPlot:
         self,
         cnv_dir,
         cnv_file,
-        genes=IMPORTANT_GENES,
+        genes=CONFIG["genes"]["default_genes_list"],
         host="localhost",
         port=8050,
     ):
@@ -453,7 +474,7 @@ class CNVPlot:
             cnv_dir (str): Path to the directory containing CNV data.
             cnv_file (str): Name of the CNV file in the directory.
             genes (list, optional): List of genes to highlight. Defaults to
-                IMPORTANT_GENES.
+                CONFIG["genes"]["default_genes_list"].
             host (str, optional): Host address for running the Dash app.
                 Defaults to "localhost".
             port (int, optional): Port number for running the Dash app.
@@ -492,14 +513,12 @@ class CNVPlot:
                 dcc.Graph(
                     id="graph",
                     figure=fig,
-                    config=dict(
-                        {
-                            "scrollZoom": False,
-                            "doubleClick": "reset",
-                            "modeBarButtonsToRemove": ["lasso2d", "select"],
-                        },
-                        displaylogo=False,
-                    ),
+                    config={
+                        "scrollZoom": False,
+                        "doubleClick": "reset",
+                        "modeBarButtonsToRemove": ["lasso2d", "select"],
+                        "displaylogo": False,
+                    },
                     style={"width": "80hh", "height": "90vh"},
                 ),
             ]
@@ -590,7 +609,8 @@ def get_cn_summary(cnv_dir, sample_ids):
     plot of gain/loss ratios across chromosomes.
 
     Args:
-        cnv_dir (str): The directory path where CNV data files are stored.
+        cnv_dir (str or Path): The directory path where CNV data files are
+            stored.
         sample_ids (list of str): A list of sample IDs to process.
 
     Returns:
