@@ -5,28 +5,35 @@ miscellaneous operations.
 """
 
 from .files import (
-    MEPYLOME_TMP_DIR,
     download_file,
     ensure_directory_exists,
     get_csv_file,
     get_file_object,
     get_resource_path,
     reset_file,
+    setup_tutorial_files,
+)
+from .varia import (
+    CONFIG,
+    MEPYLOME_TMP_DIR,
+    Timer,
+    get_free_port,
+    make_log_file,
+    normexp_get_xs,
 )
-from .tutorial_setup import setup_tutorial_files
-from .varia import Timer, get_free_port, log, normexp_get_xs
 
 __all__ = [
     "download_file",
     "ensure_directory_exists",
     "get_file_object",
     "get_csv_file",
-    "log",
+    "make_log_file",
     "reset_file",
     "get_resource_path",
     "setup_tutorial_files",
     "get_free_port",
     "Timer",
     "normexp_get_xs",
+    "CONFIG",
     "MEPYLOME_TMP_DIR",
 ]
